#ifndef GLASGOW_SUBGRAPH_SOLVER_GUARD_SRC_VERTEX_TO_VERTEX_MAPPING_HH
#define GLASGOW_SUBGRAPH_SOLVER_GUARD_SRC_VERTEX_TO_VERTEX_MAPPING_HH 1

#include <map>

namespace gss
{
    using VertexToVertexMapping = std::map<int, int>;
}

#endif
